local L = LibStub("AceLocale-3.0"):GetLocale("AffDotsDruid")

-- Damage multiplier for the various buffs
local const = {
    ["tf"]             = 1.15, -- Tigers Fury
    ["sr"]             = 1.4,  -- Savage Roar
    ["savagery_glyph"] = 1.3,  -- SR damage multiplier with Savagery Glyph 
    ["savage_roar"]    = 1.4,  -- SR damage multiplier without Savagery Glyph
    ["bt"]             = 1.3,  -- Bloodtalons 
    ["improved_rake"]  = 2,
}

-- Spell & Aura ids
local spell_id = {
    ["rip"]           = 1079,
    ["rake"]          = 155722,  -- Rake dot
	["rake_cast"]	  = 1822,    -- Rake spell
    ["improved_rake"] = 157276,  
    ["thrash"]        = 106830,
    ["shred"]         = 5221,
    ["ravage"]        = 6785,
    ["pounce"]        = 33876,
    ["swipe"]         = 33876,
    ["maim"]          = 33876,
    ["prowl"]         = 5215,
    ["shadowmeld"]    = 58984,
    ["incarnation"]   = 102543,
    
    ["fb"]            = 22568,
    ["tf"]            = 5217,
    ["sr"]            = 52610,
    ["bt"]            = 145152,  
	["sr_prowl"]      = 174544,      -- Spell id for Savage Roar applied through "Glyph of Savage Roar"
    ["sr_glyph"]      = 110816,      -- Glyph of Savagery
}

local options
local playerGuid, spec
local playerLevel                = UnitLevel("player")
local ripDps, rakeDps, thrashDps = 0, 0, 0
local timer, throttle            = 0, 0.1
local dot_damage, targets        = AffDots.dot_damage, AffDots.targets
dot_damage[spell_id.rip]         = {}
dot_damage[spell_id.rake]        = {}
dot_damage[spell_id.thrash]      = {}
local tf, sr, bt, ir             = 1, 1, 1, 1
local colorSwitch                = 1
local protect                    = 0
local bossBuffs                  = 1
local sr_cp                      = 0
local prowl_timeout              = 0
local bt_timeout                 = 0
local hasIncarnation             = false

-- Save frequently used globals as locals for performance reasons
local UnitAttackPower  = UnitAttackPower
local GetComboPoints   = GetComboPoints
local GetTime          = GetTime
local UnitLevel        = UnitLevel
local UnitGUID         = UnitGUID
local UnitDebuff       = UnitDebuff
local math_floor       = math.floor
local math_abs         = math.abs
local max              = max
local pairs            = pairs
local AffDots          = AffDots
local TidyPlates       = TidyPlates
local UnitBuff         = UnitBuff

local _, _, rake_icon    = GetSpellInfo(spell_id.rake)
local _, _, rip_icon     = GetSpellInfo(spell_id.rip)
local _, _, thrash_icon  = GetSpellInfo(spell_id.thrash)
local tf_name            = GetSpellInfo(spell_id.tf)
local sr_name            = GetSpellInfo(spell_id.sr)

local function Round(num) return math_floor(num+.5) end
local function SecondsRound(num)
    if num > 2 then return math_floor(num+.5)
    else return math_floor(num*10+.5)/10 end
end

-- Rip damage formula
local function calcRipDpsDisplay(ap, tf, sr, bt, cp)
    local damage
    if cp == 0 then
        damage = 0
    else
        damage = 20 * (sr*tf*bt*cp)
    end
    return damage
end

-- Rake damage formula
local function calcRakeDpsDisplay(ap, tf, sr, bt)
    local damage = 100 * (sr*tf*bt*ir)
	return damage
end

-- Thrash damage formula
local function calcThrashDpsDisplay(ap, tf, sr, bt)
    local damage = 100 * (sr*tf*bt)
    return damage
end

local function getDamageStats(unit)
    unit = unit or "target"
    local base, posBuff, negBuff = UnitAttackPower("player")
    local ap                     = base + posBuff + negBuff
    local cp                     = GetComboPoints("player")
    return ap, tf, sr, bt, cp
end

local function onUpdate(self,elapsed)
    timer = timer + elapsed;
    if timer >= throttle then
		for k,v in pairs(AffDots.track) do
			AffDots.track[k].update(AffDots.track[k])
			AffDots:FrameDraw(AffDots.track[k])
		end

		if options.tidyplates.enable and TidyPlates then
			AffDots:Tidy()
		end

        timer = 0
    end
end

AffDots.Update = {
    dot = function(f)
        local guid = UnitGUID(f.target)
        local _,_,_,_,_,duration,expires = UnitDebuff(f.target,f.name,f.rank,"player")
        local datanew = f.datanew
        if duration and guid and targets[guid..f.spell] then
            if(math_abs(expires-f.timer)>1) then
                datanew.expires  = expires
                datanew.duration = duration
                datanew.setcd    = true
                f.timer          = expires
            end
            datanew.color = options["color"..AffDots:FindColor(guid,expires - GetTime(),f.spell,f.pandemic)]
            datanew.tick  = targets[guid..f.spell][3]
            datanew.t2    = SecondsRound(f.timer-GetTime())
            if options.reltext then
                datanew.t1 = Round(dot_damage[f.spell][1]*100/targets[guid..f.spell][1])
            elseif options.layout == "bars" then
                datanew.t1 = dot_damage[f.spell][1].."/"..targets[guid..f.spell][1]
            else
                datanew.t1 = dot_damage[f.spell][1].."\n"..targets[guid..f.spell][1]
            end
        else
            datanew.color  = options.color7
            datanew.t1     = 0
            datanew.t2     = ""
            datanew.tick   = 0
            timer          = 0
            datanew.hidecd = true
        end
    end,
    sr = function(f)
        local _,_,_,_,_,duration,expires = UnitBuff("player", f.name)
        local datanew = f.datanew
        local refreshSR
        if duration then
            local remaining = SecondsRound(f.timer-GetTime())
            refreshSR = 5.4 + (sr_cp-1) * 1.8

            datanew.t2 = remaining
            datanew.tick  = refreshSR
            datanew.t1 = ""
            if remaining < refreshSR then
                datanew.color = options.color1
            else
                datanew.color = options.color6
            end
            if(math_abs(expires-f.timer)>1) then
                datanew.expires  = expires
                datanew.duration = duration
                datanew.setcd    = true
                f.timer          = expires
            end
        else
            datanew.color  = options.color7
            datanew.t1     = ""
            datanew.t2     = ""
            datanew.tick   = refreshSR
            f.timer        = 0
            datanew.hidecd = true
        end
    end,
}

local GetDotDmg = {
    function()
    end,
    function(self, event, unit, ...)
        -- from all tracked events the first parameter is either unit or nil
        if unit and unit ~= "player" then return end
        local ap, tf, sr, bt, cp = getDamageStats()
        -- Rip
        dot_damage[spell_id.rip][1]   = calcRipDpsDisplay(ap, tf, sr, bt, cp)
        dot_damage[spell_id.rip][2]   = 2
        -- Rake
        dot_damage[spell_id.rake][1]   = calcRakeDpsDisplay(ap, tf, sr, bt, cp)
        dot_damage[spell_id.rake][2]   = 3
        -- Thrash
        dot_damage[spell_id.thrash][1] = calcThrashDpsDisplay(ap, tf, sr, bt, cp)
        dot_damage[spell_id.thrash][2] = 3

    end,
    function() end,
    function() end,
}

local CombatLog = {
    function() end,
    function(_,_,_,event_type,_,source_GUID,_,_,_,dest_GUID,_,_,_, ...)
        if source_GUID ~= playerGuid then return end
        local spell = ...
        local t = GetTime()
        if event_type == "SPELL_AURA_APPLIED" then
            if spell == spell_id.rip then
                targets[dest_GUID..spell_id.rip]    = targets[dest_GUID..spell_id.rip] or {}
                targets[dest_GUID..spell_id.rip][1] = ripDps
                targets[dest_GUID..spell_id.rip][2] = GetTime()
                targets[dest_GUID..spell_id.rip][3] = dot_damage[spell_id.rip][2]
            elseif spell == spell_id.rake then
                rakeDps = calcRakeDpsDisplay(getDamageStats())
                if t - prowl_timeout < 0.5 and not hasIncarnation then
                    rakeDps = rakeDps * const.improved_rake
                end
                if t - bt_timeout < 0.5 then
                    rakeDps = rakeDps * const.bt
                end
                targets[dest_GUID..spell_id.rake]    = targets[dest_GUID..spell_id.rake] or {}
                targets[dest_GUID..spell_id.rake][1] = rakeDps
                targets[dest_GUID..spell_id.rake][2] = GetTime()
                targets[dest_GUID..spell_id.rake][3] = dot_damage[spell_id.rake][2]
            elseif spell == spell_id.thrash then
                thrashDps = calcThrashDpsDisplay(getDamageStats())
                targets[dest_GUID..spell_id.thrash]    = targets[dest_GUID..spell_id.thrash] or {}
                targets[dest_GUID..spell_id.thrash][1] = thrashDps
                targets[dest_GUID..spell_id.thrash][2] = GetTime()
                targets[dest_GUID..spell_id.thrash][3] = dot_damage[spell_id.thrash][2]
            elseif spell == spell_id.tf then
                tf = const.tf
            elseif spell == spell_id.sr or spell == spell_id.sr_prowl then
                sr = const.sr
                if spell == spell_id.sr_prowl then
                    sr_cp = 5
                end
            elseif spell == spell_id.bt then
                bt = const.bt
            elseif spell == spell_id.prowl or spell == spell_id.shadowmeld then
                ir = const.improved_rake
            elseif spell == spell_id.incarnation then
                ir = const.improved_rake
                hasIncarnation = true
            end
        elseif event_type == "SPELL_AURA_REFRESH" then
            if spell == spell_id.rip and GetTime() - protect > 0.5 then
                targets[dest_GUID..spell_id.rip]    = targets[dest_GUID..spell_id.rip] or {}
                targets[dest_GUID..spell_id.rip][1] = ripDps
                targets[dest_GUID..spell_id.rip][2] = GetTime()
                targets[dest_GUID..spell_id.rip][3] = dot_damage[spell_id.rip][2]
            elseif spell == spell_id.rake then
                rakeDps = calcRakeDpsDisplay(getDamageStats())
                if t - prowl_timeout < 0.5 and not hasIncarnation then
                    rakeDps = rakeDps * const.improved_rake
                end
                if t - bt_timeout < 0.5 then
                    rakeDps = rakeDps * const.bt
                end
                targets[dest_GUID..spell_id.rake]    = targets[dest_GUID..spell_id.rake] or {}
                targets[dest_GUID..spell_id.rake][1] = rakeDps
                targets[dest_GUID..spell_id.rake][2] = GetTime()
                targets[dest_GUID..spell_id.rake][3] = dot_damage[spell_id.rake][2]
            elseif spell == spell_id.thrash then
                thrashDps = calcThrashDpsDisplay(getDamageStats())
                targets[dest_GUID..spell_id.thrash]    = targets[dest_GUID..spell_id.thrash] or {}
                targets[dest_GUID..spell_id.thrash][1] = thrashDps
                targets[dest_GUID..spell_id.thrash][2] = GetTime()
                targets[dest_GUID..spell_id.thrash][3] = dot_damage[spell_id.thrash][2]
            end
        elseif event_type == "SPELL_DAMAGE" then
            if (spell == spell_id.shred or spell == spell_id.fb or spell == spell_id.ravage) then
                protect = GetTime()
            end
        elseif event_type == "SPELL_CAST_SUCCESS" then
            if spell == spell_id.rip then
                ripDps = calcRipDpsDisplay(getDamageStats())
            elseif spell == spell_id.sr then
                sr_cp = GetComboPoints("player")
            elseif spell == spell_id.rake_cast then
            end
        elseif event_type == "SPELL_AURA_REMOVED" then
            if spell == spell_id.tf then
                tf = 1
            elseif spell == spell_id.bt then
                bt = 1
                bt_timeout = GetTime()
            elseif spell == spell_id.sr or spell == spell_id.sr_prowl then
                sr = 1
            elseif (spell == spell_id.prowl or spell == spell_id.shadowmeld) and not hasIncarnation then
                prowl_timeout = GetTime()
                ir = 1
            elseif spell == spell_id.incarnation then
                ir = 1
                hasIncarnation = false
            end
        end
    end,
    function() end,
    function() end,
}

local function hasSavageryGlyph() 
    local glyphCount = GetNumGlyphSockets()
    local i
    for i = 0, glyphCount do
        local enabled, glyphType, glyphTooltipIndex, glyphSpellID, icon = GetGlyphSocketInfo(i)
        if enabled and glyphSpellID == const.sr_glyph then
            return true
        end
    end
    return false
end

function AffDots:InitSpec()
    AffDots:ReleaseFrames()
    wipe(AffDots.track)
    if spec == 2 then
        AffDots.execute_percent = 25
        AffDots:Track( spell_id.rip,    AffDots.Update.dot, "target", 24, 2, options.feral.rip_order,    options.feral.show_rip)
        AffDots:Track( spell_id.rake,   AffDots.Update.dot, "target", 15, 3, options.feral.rake_order,   options.feral.show_rake)
        AffDots:Track( spell_id.thrash, AffDots.Update.dot, "target", 15, 3, options.feral.thrash_order, options.feral.show_thrash)
        AffDots:Track( spell_id.sr,     AffDots.Update.sr,  "target", 42, 0, options.feral.sr_order,     options.feral.show_sr)
        
        if options.focus then
            AffDots:Track( spell_id.rip,    AffDots.Update.dot, "focus", 24, 2, options.feral.rip_order,    options.feral.show_rip)
            AffDots:Track( spell_id.rake,   AffDots.Update.dot, "focus", 15, 3, options.feral.rake_order,   options.feral.show_rake)
            AffDots:Track( spell_id.thrash, AffDots.Update.dot, "focus", 15, 3, options.feral.thrash_order, options.feral.show_thrash)
        end
		AffDots.tidyspells = {
            [rake_icon]   = {spell = spell_id.rake,   pandemic = 4.5},
            [rip_icon]    = {spell = spell_id.rip,    pandemic = 7.2},
            [thrash_icon] = {spell = spell_id.thrash, pandemic = 4.5},
        }
    end
    colorSwitch = options.feral.color_switch
    if hasSavageryGlyph() then const.sr = const.savagery else const.sr = const.savage_roar end
    if playerLevel < 98 then
        const.improved_rake = 1
    end
    AffDots:InitFrames()
end

function AffDots:InitializeClass()
    playerGuid = UnitGUID("player")
    AffDots:RegisterEvent("PLAYER_TALENT_UPDATE", "InitializeClass")
    AffDots:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
--    AffDots:UnregisterEvent("COMBAT_RATING_UPDATE")
--    AffDots:UnregisterEvent("SPELL_POWER_CHANGED")
--    AffDots:UnregisterEvent("PLAYER_DAMAGE_DONE_MODS")
--    AffDots:UnregisterEvent("UNIT_STATS")
--    AffDots:UnregisterEvent("UNIT_ATTACK_POWER")
    AffDots:UnregisterEvent("UNIT_COMBO_POINTS")
    AffDots:UnregisterEvent("UNIT_AURA")
    AffDots:UnregisterEvent("PLAYER_REGEN_ENABLED")
    AffDots:UnregisterEvent("PLAYER_REGEN_DISABLED")
    AffDots:UnregisterEvent("PLAYER_FOCUS_CHANGED")
    -- New for 6.02
    AffDots:UnregisterEvent("GLYPH_ADDED")
    AffDots:UnregisterEvent("GLYPH_REMOVED")
    --
    AffDotsTarget:SetScript("OnUpdate", nil)
    AffDotsTarget:Hide(); AffDotsFocus:Hide();
    spec = GetSpecialization() or ""
    if (spec == 2) then
        options                    = self.db.profile
        options.feral              = options.feral or {}
        options.feral.rip_order    = options.feral.rip_order or 0
        options.feral.rake_order   = options.feral.rake_order or 1
        options.feral.thrash_order = options.feral.thrash_order or 2
        options.feral.sr_order     = options.feral.sr_order or 3
        options.feral.color_switch = options.feral.color_switch or 1
        options.feral.show_rip     = options.feral.show_rip == nil or options.feral.show_rip
        options.feral.show_rake    = options.feral.show_rake == nil or options.feral.show_rake
        options.feral.show_thrash  = options.feral.show_thrash == nil or options.feral.show_thrash
        options.feral.show_sr      = options.feral.show_sr == nil or options.feral.show_sr
        colorSwitch                = options.feral.color_switch

        AffDots.GetDotDmg = GetDotDmg[spec]
        AffDots.CombatLog = CombatLog[spec]
        AffDots:GetDotDmg()
        AffDots:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "CombatLog")
--        AffDots:RegisterEvent("COMBAT_RATING_UPDATE", "GetDotDmg")
--        AffDots:RegisterEvent("SPELL_POWER_CHANGED", "GetDotDmg")
--        AffDots:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "GetDotDmg")
--        AffDots:RegisterEvent("UNIT_STATS", "GetDotDmg")
--        AffDots:RegisterEvent("UNIT_ATTACK_POWER", "GetDotDmg")
        AffDots:RegisterEvent("UNIT_COMBO_POINTS", "GetDotDmg")
        AffDots:RegisterEvent("UNIT_AURA", "GetDotDmg")
        -- New for 6.02
        AffDots:RegisterEvent("GLYPH_ADDED", "InitSpec")
        AffDots:RegisterEvent("GLYPH_REMOVED", "InitSpec")
        --
        AffDots:RegisterEvent("PLAYER_REGEN_ENABLED")
        AffDots:RegisterEvent("PLAYER_REGEN_DISABLED")
        AffDots:RegisterEvent("PLAYER_FOCUS_CHANGED")
        AffDotsTarget:SetScript("OnUpdate", onUpdate)
        AffDots:InitSpec()
    end
end

function AffDots:GetSpecOptions(spec)
    if spec == "DRUID2" then
        self.options.args.spec.name = L["Feral"]
        self.options.args.spec.args = {
			show_rip = {
				order = 1, name = L["Show Rip"], type = "toggle",
				set = function(info,val) options.feral.show_rip = not options.feral.show_rip; AffDots:InitSpec() end,
				get = function(info) return options.feral.show_rip end
			},
			show_rake = {
				order = 2, name = L["Show Rake"], type = "toggle",
				set = function(info,val) options.feral.show_rake = not options.feral.show_rake; AffDots:InitSpec() end,
				get = function(info) return options.feral.show_rake end
			},
			show_thrash = {
				order = 3,name = L["Show Thrash"],type = "toggle",
				set = function(info,val) options.feral.show_thrash = not options.feral.show_thrash; AffDots:InitSpec() end,
				get = function(info) return options.feral.show_thrash end
			},
			show_sr = {
				order = 3,name = L["Show Savage Roar"],type = "toggle",
				set = function(info,val) options.feral.show_sr = not options.feral.show_sr; AffDots:InitSpec() end,
				get = function(info) return options.feral.show_sr end
			},
            desc1 = {
                order = 4,name = L["Order:"],type = "description",fontSize="medium"
            },
			rip_order = {
				order = 5,name = L["Rip"],type = "range",
				min=1, max=4, step=1, softMin=1, softMax=4,
				set = function(info,val) options.feral.rip_order = val - 1; AffDots:InitSpec() end,
				get = function(info) return options.feral.rip_order + 1 end
			},
			rake_order = {
				order = 6,name = L["Rake"],type = "range",
				min=1, max=4, step=1, softMin=1, softMax=4,
				set = function(info,val) options.feral.rake_order = val - 1; AffDots:InitSpec() end,
				get = function(info) return options.feral.rake_order + 1 end
			},
			thrash_order = {
				order = 6, name = L["Thrash"], type = "range",
				min=1, max=4, step=1, softMin=1, softMax=4,
				set = function(info,val) options.feral.thrash_order = val - 1; AffDots:InitSpec() end,
				get = function(info) return options.feral.thrash_order + 1 end
			},
			sr_order = {
				order = 7, name = L["Savage Roar"], type = "range",
				min=1, max=4, step=1, softMin=1, softMax=4,
				set = function(info,val) options.feral.sr_order = val - 1; AffDots:InitSpec() end,
				get = function(info) return options.feral.sr_order + 1 end
			},
        }

        self.options.args.colors.args={
            color1 = {
                order = 1,width = "full",name = L["Current stats are better, full Pandemic benefit"],type = "color",
                set = function(info,r,g,b) self.db.profile.color1={r,g,b}; self:SaveColors() end,
                get = function(info) return self.db.profile.color1[1],self.db.profile.color1[2],self.db.profile.color1[3],1 end
            },
            color2 = {
                order = 2,width = "full",name = L["Current stats are same, full Pandemic benefit"],type = "color",
                set = function(info,r,g,b) self.db.profile.color2={r,g,b}; self:SaveColors()  end,
                get = function(info) return self.db.profile.color2[1],self.db.profile.color2[2],self.db.profile.color2[3],1 end
            },
            color3 = {
                order = 3,width = "full",name = L["Current stats are worse, full Pandemic benefit"],type = "color",
                set = function(info,r,g,b) self.db.profile.color3={r,g,b}; self:SaveColors()  end,
                get = function(info) return self.db.profile.color3[1],self.db.profile.color3[2],self.db.profile.color3[3],1 end
            },
            color4 = {
                order = 4,width = "full",name = L["Current stats are better, partial Pandemic benefit"],type = "color",
                set = function(info,r,g,b) self.db.profile.color4={r,g,b}; self:SaveColors()  end,
                get = function(info) return self.db.profile.color4[1],self.db.profile.color4[2],self.db.profile.color4[3],1 end
            },
            color5 = {
                order = 5,width = "full",name = L["Current stats are same, partial Pandemic benefit"],type = "color",
                set = function(info,r,g,b) self.db.profile.color5={r,g,b}; self:SaveColors()  end,
                get = function(info) return self.db.profile.color5[1],self.db.profile.color5[2],self.db.profile.color5[3],1 end
            },
            color6 = {
                order = 6,width = "full",name = L["Current stats are worse, partial Pandemic benefit"],type = "color",
                set = function(info,r,g,b) self.db.profile.color6={r,g,b}; self:SaveColors()  end,
                get = function(info) return self.db.profile.color6[1],self.db.profile.color6[2],self.db.profile.color6[3],1 end
            },
            color7 = {
                order = 7,width = "full",name = L["No dot on target"],type = "color",
                set = function(info,r,g,b) self.db.profile.color7={r,g,b}; self:SaveColors()  end,
                get = function(info) return self.db.profile.color7[1],self.db.profile.color7[2],self.db.profile.color7[3],1 end
            },            
            text_color = {
                order  = 8, width = "full", name = L["Text color"], type = "color",
                set    = function(info,r,g,b) options.color11={r,g,b} end,
                get    = function(info) return options.color11[1], options.color11[2], options.color11[3],1 end
            },
            color_switch = {
                order = 9, name = L["Color switch"], type = "range",
                desc  = L["Switch color only if current stats are at least x% better"],
                min   = 0, max = 40, step = 1, softMin = 0, softMax=40,
                set   = function(info,val) options.feral.color_switch = 1+val/100 end,
                get   = function(info) return (options.feral.color_switch-1)*100 end
            },
        }
    else
        self.options.args.spec.name = "Specialization"
        self.options.args.spec.args = {
            spacer1 = {
                order = 1,type = "description",name=L["Specialization not supported, AffDots deactivated"], width="full",
            },
        }
    end
end

function AffDots:FindColor(GUID,time_left,spell,pandemic)
    if not targets[GUID..spell] or not dot_damage[spell] then
        return 7
    end
    local  current   = dot_damage[spell][1]
    local  on_target = targets[GUID..spell][1]
    local color = 5

	if		(on_target*colorSwitch < current and time_left <= pandemic) then color = 1
	elseif 	(on_target == current and time_left <= pandemic)            then color = 2
	elseif	(on_target  > current and time_left <= pandemic)            then color = 3
	elseif	(on_target*colorSwitch < current and time_left  > pandemic) then color = 4
	elseif	(on_target  > current and time_left  > pandemic)            then color = 6
	end
--[[    
    if on_target*colorSwitch < current then
        color = 1
    elseif on_target > current then
        color = 6
    else
        color = 2
    end
]]
    return color
end

-- Fix for not being able to detect talent switches properly because spec is a local variable in AffDots
function AffDotsHandler(msg, editBox)
	AffDots:GetLayoutOptions()
	AffDots:GetSpecOptions("DRUID"..spec)
	LibStub("AceConfigDialog-3.0"):Open("AffDots")
end
SlashCmdList["AFFDOTS"] = AffDotsHandler
